import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../models/models.dart';

abstract class SearchFormTypeCollection {
  const SearchFormTypeCollection();
}

class SearchClause extends Equatable {
  final M4eUniqueId subjectId;
  final M4eAccessPermissionType type;
  final List<M4eAccessPermissionSet> encoding;

  const SearchClause._(this.subjectId, this.type, this.encoding);

  factory SearchClause({
    @required M4eUniqueId subjectId,
    @required M4eAccessPermissionType type,
    @required List<M4eAccessPermissionSet> encoding,
  }) {
    assert(subjectId != null,
        'SingleClauseSearchForm [subjectId] should not be null');
    assert(type != null, 'SingleClauseSearchForm [type] should not be null');
    assert(encoding != null && encoding.isNotEmpty,
        'SingleClauseSearchForm [encoding] should not be null');

    return SearchClause._(subjectId, type, encoding);
  }

  Map<String, dynamic> toJson() {
    return {
      'subjectId': this.subjectId.toString(),
      'encoding': {
        'type': this.type.toString(),
        'encoding': this.encoding?.map((e) => e.toString())?.toList() ?? [],
      }
    };
  }

  @override
  List<Object> get props => [subjectId, type, encoding];

  @override
  String toString() =>
      'SearchClause(subjectId: $subjectId, type: ${type.toString()}, encoding: ${encoding.toString()})';
}

// SINGLE CLAUSE SEARCH FORM ===========================================================================================================
class SingleClauseSearchForm extends SearchFormTypeCollection
    with EquatableMixin {
  final M4eUniqueId id;
  final SearchClause clause;

  const SingleClauseSearchForm._(this.id, this.clause);

  factory SingleClauseSearchForm({@required SearchClause clause}) {
    assert(
        clause != null, 'SingleClauseSearchForm [clause] should not be null');

    return SingleClauseSearchForm._(M4eUniqueId(), clause);
  }

  Map<String, dynamic> toJson() {
    return {
      'id': this.id.toString(),
      'clause': this.clause.toJson(),
    };
  }

  //? omitted `id` field since its auto generated wouldnt be ideal for value comparism
  @override
  List<Object> get props => [clause];
  // List<Object> get props => [id, clause];

  @override
  String toString() => 'SingleClauseSearchForm(id: $id, clause: $clause)';
}

// MULTIPLE CLAUSE SEARCH FORM ===========================================================================================================
class MultipleClauseSearchForm extends SearchFormTypeCollection
    with EquatableMixin {
  final M4eUniqueId id;
  final M4eSearchConjunctionType conjunction;
  final List<SearchClause> clauses;

  const MultipleClauseSearchForm._(this.id, this.conjunction, this.clauses);

  factory MultipleClauseSearchForm({
    @required M4eSearchConjunctionType conjunction,
    @required List<SearchClause> clauses,
  }) {
    assert(conjunction != null,
        'MultipleClauseSearchForm [conjunction] should not be null');
    assert(clauses != null && clauses.isNotEmpty,
        'MultipleClauseSearchForm [clauses] should not be null');

    return MultipleClauseSearchForm._(M4eUniqueId(), conjunction, clauses);
  }

  Map<String, dynamic> toJson() {
    return {
      'id': this.id.toString(),
      'clause': {
        'conjunction': this.conjunction.toString(),
        'clauses': this.clauses?.map((e) => e.toJson())?.toList() ?? []
      }
    };
  }

  //? omitted `id` field since its auto generated wouldnt be ideal for value comparism
  @override
  List<Object> get props => [conjunction, clauses];
  // List<Object> get props => [id, conjunction, clauses];

  @override
  String toString() =>
      'MultipleClauseSearchForm(id: $id, conjunction: $conjunction, clauses: $clauses)';
}

class MobileMoneyProviderSearchForm extends SearchFormTypeCollection
    with EquatableMixin {
  final M4eUniqueId id;
  final M4eUniqueId objectId;
  final M4eWalletProviderType provider;
  final M4ePhoneNumber phoneNumber;

  const MobileMoneyProviderSearchForm._(
      this.id, this.objectId, this.provider, this.phoneNumber);

  factory MobileMoneyProviderSearchForm(
      {@required M4eWalletProviderType provider,
      @required M4ePhoneNumber phoneNumber}) {
    assert(provider != null,
        'MobileMoneyProviderSearchForm [provider] should not be null');
    assert(phoneNumber != null,
        'MobileMoneyProviderSearchForm [phoneNumber] should not be null');

    return MobileMoneyProviderSearchForm._(
        M4eUniqueId(), M4eUniqueId(), provider, phoneNumber);
  }

  Map<String, dynamic> toJson() {
    return {
      'id': this.id.toString(),
      'clause': {
        'objectId': this.objectId.toString(),
        'detail': {
          'provider': this.provider.toString(),
          'phone': this.phoneNumber.toJson(),
        }
      }
    };
  }

  //? omitted `id` field since its auto generated wouldnt be ideal for value comparism
  @override
  List<Object> get props => [provider, phoneNumber];
  // List<Object> get props => [id, objectId, provider, phoneNumber];

  @override
  String toString() {
    return 'MobileMoneyProviderSearchForm(id: $id, objectId: $objectId, provider: $provider, phoneNumber: $phoneNumber)';
  }
}
