import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';

import '../../core/core.dart';
import '../../models/models.dart';

/// Form parameters required for m4e-sdk transactions
///
class M4eTransactionForm extends Equatable {
  final M4eWallet from;
  final M4eWallet to;
  final M4eAmount amount;
  final DateTime at;
  final Map<String, dynamic> meta;

  const M4eTransactionForm._({
    @required this.from,
    @required this.to,
    @required this.amount,
    @required this.at,
    @required this.meta,
  });

  factory M4eTransactionForm({
    @required M4eWallet from,
    @required M4eWallet to,
    @required M4eAmount amount,
    @required DateTime at,
    @required Map<String, dynamic> meta,
  }) {
    assert(from != null, 'M4eTransactionForm [from] wallet should not be null');
    assert(to != null, 'M4eTransactionForm [to] wallet should not be null');
    assert(from.id != to.id,
        'M4eTransactionForm [from] and [to] wallets should not be the same');
    assert(amount != null, 'M4eTransactionForm [amount] should not be null');
    assert(at != null, 'M4eTransactionForm [at] should not be null');
    assert(meta != null, 'M4eTransactionForm [meta] should not be null');

    return M4eTransactionForm._(
        from: from, to: to, amount: amount, at: at, meta: meta);
  }

  factory M4eTransactionForm.fromJson(Map<String, dynamic> json) {
    return M4eTransactionForm._(
      from: M4eWallet.fromJson(json['flow']['from']),
      to: M4eWallet.fromJson(json['flow']['to']),
      amount: M4eAmount.fromJson(json['flow']['from']),
      at: DateTimeConverter.fromTimestampSinceEpoch(json['flow']['from']),
      meta: json['meta'],
    );
  }

  M4eTransactionForm copyWith({
    M4eWallet from,
    M4eWallet to,
    M4eAmount amount,
    DateTime at,
    Map<String, dynamic> meta,
  }) {
    return M4eTransactionForm(
      from: from ?? this.from,
      to: to ?? this.to,
      amount: amount ?? this.amount,
      at: at ?? this.at,
      meta: meta ?? this.meta,
    );
  }

  Map<String, dynamic> toJson() {
    final flow = {
      'from': this.from.toJson(),
      'to': this.to.toJson(),
      'amount': this.amount.toJson(),
      'at': DateTime.now().toTimestampSinceEpoch()
    };

    // TODO: Transaction is a test object. Remove when not needed
    final _transactionType =
        this.from.walletInfo.type == M4eWalletProviderType.M4E &&
                this.to.walletInfo.type == M4eWalletProviderType.M4E
            ? M4eTransactionType.cashIn
            : M4eTransactionType.cashOut;

    return {
      'flow': flow,
      'txType': _transactionType.toString(),
      'meta': this.meta,
    };
  }

  @override
  List<Object> get props => [
        from,
        to,
        amount,
        at,
        meta,
      ];

  @override
  String toString() {
    return 'M4eTransactionForm(from: $from, to: $to, amount: $amount, at: $at, meta: $meta)';
  }
}
