import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';

import '../core/core.dart';
import '../models/models.dart';
import '../m4e_forms/m4e_forms.dart';

class M4eWalletInfo extends Equatable {
  final String label;
  final M4eCurrencyType currency;
  final M4eWalletProviderType type;

  const M4eWalletInfo._(this.label, this.currency, this.type);

  factory M4eWalletInfo({
    @required String label,
    @required M4eCurrencyType currency,
    @required M4eWalletProviderType type,
  }) {
    assert(label != null, 'WalletInfo [label] should not be null');
    assert(currency != null, 'WalletInfo [currency] should not be null');
    assert(type != null, 'WalletInfo [type] should not be null');

    return M4eWalletInfo._(label, currency, type);
  }

  factory M4eWalletInfo.fromJson(Map<String, dynamic> json) {
    return M4eWalletInfo(
      label: json['label'],
      currency: M4eCurrencyType.parse(json['currency']),
      type: M4eWalletProviderType.parse(json['type']),
    );
  }

  @override
  List<Object> get props => [
        label,
        currency,
        type,
      ];

  @override
  String toString() =>
      'M4eWalletInfo(label: $label, currency: ${currency.toString()}, type: $type)';
}

class M4eWallet extends Equatable {
  final M4eUniqueId id;
  final M4eWalletInfo walletInfo;
  final M4eWalletDetailType detail;

  const M4eWallet._(this.id, this.walletInfo, this.detail);

  factory M4eWallet({
    @required M4eUniqueId id,
    @required M4eWalletInfo walletInfo,
    @required M4eWalletDetailType detail,
  }) {
    assert(id != null, 'Wallet [id] should not be null');
    assert(walletInfo != null, 'Wallet [walletInfo] should not be null');
    assert(detail != null, 'M4eWalletForm [detail] should not be null');

    // TODO: consider validate [M4eWalletProviderType] to its detail-id

    // validate [M4eWalletProviderType] to [M4eWalletDetailType] relationship
    if (walletInfo.type == M4eWalletProviderType.M4E &&
        detail is MobileMoneyWalletDetail) {
      // TODO: update [M4eWalletProviderType] to M4E from BASCULE
      throw AssertionError(
          'M4eWalletForm [detail] should not be of type [OtherWalletDetail] when [M4eWalletProviderType] is BASCULE! \nUse [M4eWalletDetail] as [M4eWalletDetailType] type instead');
    }

    return M4eWallet._(id, walletInfo, detail);
  }

  M4eWallet copyWith(
      {M4eUniqueId id, M4eWalletInfo walletInfo, M4eWalletDetailType detail}) {
    return M4eWallet(
      id: id ?? this.id,
      walletInfo: walletInfo ?? this.walletInfo,
      detail: detail ?? this.detail,
    );
  }

  /// transforms [M4eWallet] to [M4eWalletForm] type
  M4eWalletForm toWalletFormType() {
    return M4eWalletForm.fromJson(this.toJson());
  }

  factory M4eWallet.fromJson(Map<String, dynamic> json) {
    return M4eWallet(
      id: M4eUniqueId.fromUniqueString(json['id']),
      walletInfo: M4eWalletInfo.fromJson(json['wallet']),
      detail: M4eWalletDetailType.fromJson(json['detail']),
    );
  }

  Map<String, dynamic> toJson() {
    return {
      'id': this.id.toString(),
      'wallet': {
        'label': this.walletInfo.label,
        'currency': this.walletInfo.currency.toString(),
        'type': this.walletInfo.type.toString()
      },
      'detail': this.detail.map(
            m4eWalletDetail: (detail) => detail.toJson(),
            mobileMoneyWalletDetail: (detail) => detail.toJson(),
          ),
    };
  }

  @override
  List<Object> get props => [id, walletInfo];

  @override
  bool get stringify => true;
}

class M4eWalletList extends Equatable {
  final List<M4eWallet> wallets;

  const M4eWalletList({@required this.wallets});

  factory M4eWalletList.fromJson(Map<String, dynamic> json) {
    final _data = json['result'] as List;

    return M4eWalletList(
        wallets: _data?.map((e) => M4eWallet.fromJson(e))?.toList() ?? []);
  }

  Map<String, dynamic> toJson() {
    return {
      'result': this.wallets?.map((e) => e.toJson())?.toList() ?? [],
    };
  }

  @override
  List<Object> get props => [wallets];

  @override
  String toString() => 'M4eWalletList(wallets: $wallets)';
}
