import 'package:data_connection_checker/data_connection_checker.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:m4e_sdk/src/core/core.dart';
import 'package:m4e_sdk/src/data/api/apis.dart';
import 'package:m4e_sdk/src/data/services/services.dart';
import 'package:m4e_sdk/src/m4e_forms/m4e_forms.dart';
import 'package:m4e_sdk/src/models/models.dart';
import 'package:matcher/matcher.dart';
import 'package:mockito/mockito.dart';

import 'test_models/wallet_test_model.dart';
// import 'package:matcher/matcher.dart';

class MockWalletApi extends Mock implements WalletApi {}

class MockTransactionApi extends Mock implements TransactionApi {}

class MockDataConnectionChecker extends Mock implements DataConnectionChecker {}

main() {
  M4eTransactionService _m4eTransactionService;

  MockWalletApi _mockWalletApi;
  MockTransactionApi _mockTransactionApi;
  MockDataConnectionChecker _mockDataConnectionChecker;

  setUp(() {
    _mockWalletApi = MockWalletApi();
    _mockTransactionApi = MockTransactionApi();
    _mockDataConnectionChecker = MockDataConnectionChecker();

    _m4eTransactionService = M4eTransactionService(
      walletApi: _mockWalletApi,
      transactionApi: _mockTransactionApi,
      connectionChecker: _mockDataConnectionChecker,
    );
  });

  final _uuid =
      M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0');
  final _fromUuId =
      M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0');
  final _toUuId =
      M4eUniqueId.fromUniqueString('9e8b5ecc-9f7e-4e06-ac3d-05b11a91da88');

  final _phoneNumber = M4ePhoneNumber(number: '0541234567', country: 'GHANA');

  final _m4eWalletDetail = M4eWalletDetail(id: _toUuId);
  final _mobileMoneyWalletDetail = MobileMoneyWalletDetail(
      provider: M4eWalletProviderType.MTN, phoneNumber: _phoneNumber);

  final _tM4eToM4eTransaction = M4eTransactionForm(
    from: M4eWallet(
        id: _fromUuId,
        walletInfo: M4eWalletInfo(
          label: 'W1',
          currency: M4eCurrencyType.GHS,
          type: M4eWalletProviderType.M4E,
        ),
        detail: _m4eWalletDetail),
    to: M4eWallet(
      id: _toUuId,
      walletInfo: M4eWalletInfo(
        label: 'W2',
        currency: M4eCurrencyType.GHS,
        type: M4eWalletProviderType.M4E,
      ),
      detail: _m4eWalletDetail,
    ),
    amount: M4eAmount(currency: M4eCurrencyType.GHS, value: 50.0),
    at: DateTime.now(),
    meta: <String, dynamic>{
      'callbackUrl': 'www.google.com',
      'purpose': 'payment',
      'data': 'test data',
    },
  );

  final _tM4eToMTNTransaction = M4eTransactionForm(
    from: M4eWallet(
      id: _fromUuId,
      walletInfo: M4eWalletInfo(
        label: 'W1',
        currency: M4eCurrencyType.GHS,
        type: M4eWalletProviderType.M4E,
      ),
      detail: _m4eWalletDetail,
    ),
    to: M4eWallet(
      id: _toUuId,
      walletInfo: M4eWalletInfo(
        label: 'W2',
        currency: M4eCurrencyType.GHS,
        type: M4eWalletProviderType.MTN,
      ),
      detail: _mobileMoneyWalletDetail,
    ),
    amount: M4eAmount(currency: M4eCurrencyType.GHS, value: 50.0),
    at: DateTime.now(),
    meta: <String, dynamic>{
      'callbackUrl': 'www.google.com',
      'purpose': 'payment',
      'data': 'test data',
    },
  );

  final _tTransactionResult = M4eTransactionResult(
    from: _fromUuId,
    to: _toUuId,
    amount: M4eAmount(
      currency: M4eCurrencyType.GHS,
      value: 50.0,
    ),
    at: DateTime.now(),
    txType: M4eTransactionType.parse('CASH_IN'),
    meta: <String, dynamic>{
      'callbackUrl': 'www.google.com',
      'purpose': 'payment',
      'data': 'test data',
    },
  );

  final _userId =
      M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0');

  final _transactionLogs = [
    M4eTransaction(
      id: _uuid,
      from: M4eWallet(
        id: _fromUuId,
        walletInfo: M4eWalletInfo(
          label: 'W1',
          currency: M4eCurrencyType.GHS,
          type: M4eWalletProviderType.M4E,
        ),
        detail: _m4eWalletDetail,
      ),
      to: M4eWallet(
        id: _toUuId,
        walletInfo: M4eWalletInfo(
          label: 'W2',
          currency: M4eCurrencyType.GHS,
          type: M4eWalletProviderType.M4E,
        ),
        detail: _m4eWalletDetail,
      ),
      amount: M4eAmount(currency: M4eCurrencyType.parse('GHC'), value: 50.0),
      at: DateTime.now(),
      type: M4eTransactionType.cashIn,
      meta: <String, dynamic>{
        'callbackUrl': 'www.google.com',
        'purpose': 'payment',
        'data': 'test data',
      },
    ),
  ];

  final _transactionList = M4eTransactionList(transactions: _transactionLogs);

  group('m4e-transaction-test initiateTransaction |', () {
    test('should verify device connectivity status', () async {
      // arrange
      when(_mockDataConnectionChecker.hasConnection)
          .thenAnswer((_) async => true);
      when(_mockTransactionApi.initiateTransaction(any))
          .thenAnswer((_) async => _tTransactionResult);

      // act
      await _m4eTransactionService.initiateTransaction(
          _fromUuId, _tM4eToM4eTransaction);

      // assert
      verify(_mockDataConnectionChecker.hasConnection);
    });

    group('isConnected |', () {
      setUp(() {
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
      });

      test('should return [M4eTransactionResult] on success call', () async {
        // arrange
        when(_mockTransactionApi.initiateTransaction(any))
            .thenAnswer((_) async => _tTransactionResult);

        // act
        final _result = await _m4eTransactionService.initiateTransaction(
            _fromUuId, _tM4eToM4eTransaction);

        // assert
        expect(_result, equals(_tTransactionResult));
      });

      test(
          'should return [M4eTransactionResult] from TransactionApi on success call',
          () async {
        // arrange
        when(_mockTransactionApi.initiateTransaction(any))
            .thenAnswer((_) async => _tTransactionResult);

        // act
        await _m4eTransactionService.initiateTransaction(
            _fromUuId, _tM4eToM4eTransaction);

        // assert
        verify(_mockTransactionApi.initiateTransaction(_tM4eToM4eTransaction));
      });

      group('mobileMoneyProviderTransactionMapper', () {
        final _composedMobileMoneyWalletSearchForm =
            MobileMoneyProviderSearchForm(
          provider: _tM4eToMTNTransaction.to.walletInfo.type,
          phoneNumber: _tM4eToMTNTransaction.to.detail.maybeMap(
            mobileMoneyWalletDetail: (detail) => detail.phoneNumber,
            orElse: () => null,
          ),
        );

        final _m4eToMTNTransactionResult = M4eTransactionResult(
          from: _fromUuId,
          to: _tM4eToMTNTransaction.to.id,
          amount: _tTransactionResult.amount,
          at: _tTransactionResult.at,
          txType: M4eTransactionType.cashOut,
          meta: _tTransactionResult.meta,
        );

        test(
            'should call [WalletApi] `queryUserWallets` method with appropriate parameters ',
            () async {
          // arrange
          when(_mockWalletApi.queryUserWallets(any))
              .thenAnswer((_) async => WalletTestModel.uniqueId);
          when(_mockWalletApi.runSearchQuery(any)).thenAnswer((_) async =>
              WalletTestModel.tWalletListAnonymousSearchByPhoneNumber);

          // act
          await _m4eTransactionService.initiateTransaction(
              _userId, _tM4eToMTNTransaction);

          // assert
          verify(_mockWalletApi
              .queryUserWallets(_composedMobileMoneyWalletSearchForm));
        });

        test(
            'should call [WalletApi] `runSearchQuery` method to run search-query-id with appropriate parameters ',
            () async {
          // arrange
          when(_mockWalletApi.queryUserWallets(any))
              .thenAnswer((_) async => WalletTestModel.uniqueId);
          when(_mockWalletApi.runSearchQuery(any)).thenAnswer((_) async =>
              WalletTestModel.tWalletListAnonymousSearchByPhoneNumber);

          // act
          await _m4eTransactionService.initiateTransaction(
              _userId, _tM4eToMTNTransaction);

          // assert
          verify(_mockWalletApi.runSearchQuery(WalletTestModel.uniqueId));
        });

        test(
            'should call [WalletApi] `createWallet` method to create a wallet if none was fetched from search-query',
            () async {
          // arrange
          when(_mockWalletApi.queryUserWallets(any))
              .thenAnswer((_) async => WalletTestModel.uniqueId);
          when(_mockWalletApi.runSearchQuery(any))
              .thenAnswer((_) async => WalletTestModel.tWalletListEmpty);

          // act
          await _m4eTransactionService.initiateTransaction(
              _userId, _tM4eToMTNTransaction);

          // assert
          verify(_mockWalletApi
              .queryUserWallets(_composedMobileMoneyWalletSearchForm));
          verify(_mockWalletApi.runSearchQuery(WalletTestModel.uniqueId));
          verify(_mockWalletApi.createWallet(
              _userId, _tM4eToMTNTransaction.to.toWalletFormType()));
        });

        test(
            'should update [M4eTransactionForm] `to` object to returned anounymous wallet from search',
            () async {
          // arrange
          when(_mockWalletApi.queryUserWallets(any))
              .thenAnswer((_) async => WalletTestModel.uniqueId);
          when(_mockWalletApi.runSearchQuery(any)).thenAnswer((_) async =>
              WalletTestModel.tWalletListAnonymousSearchByPhoneNumber);
          when(_mockTransactionApi.initiateTransaction(any))
              .thenAnswer((_) async => _m4eToMTNTransactionResult);

          // act
          await _m4eTransactionService.initiateTransaction(
              _userId, _tM4eToMTNTransaction);

          // assert
          verify(_mockWalletApi
              .queryUserWallets(_composedMobileMoneyWalletSearchForm));
          verify(_mockWalletApi.runSearchQuery(WalletTestModel.uniqueId));
          verify(
              _mockTransactionApi.initiateTransaction(_tM4eToMTNTransaction));
        });

        test(
            'should call [TransactionApi] to initiate transaction with appropriate parameters',
            () async {
          // arrange
          when(_mockWalletApi.queryUserWallets(any))
              .thenAnswer((_) async => WalletTestModel.uniqueId);
          when(_mockWalletApi.runSearchQuery(any))
              .thenAnswer((_) async => WalletTestModel.tWalletListEmpty);
          when(_mockWalletApi.createWallet(any, any))
              .thenAnswer((_) async => _tM4eToMTNTransaction.to.id);

          // act
          await _m4eTransactionService.initiateTransaction(
              _userId, _tM4eToMTNTransaction);

          verify(_mockWalletApi
              .queryUserWallets(_composedMobileMoneyWalletSearchForm));
          verify(_mockWalletApi.runSearchQuery(WalletTestModel.uniqueId));
          verify(_mockWalletApi.createWallet(
              _userId, _tM4eToMTNTransaction.to.toWalletFormType()));
          verify(
              _mockTransactionApi.initiateTransaction(_tM4eToMTNTransaction));

          // expect(_result, equals(_expectedResult));
        });

        test(
            'should return [TransactionResult] with `to` object as the uuid of the created wallet',
            () async {
          // arrange
          when(_mockWalletApi.queryUserWallets(any))
              .thenAnswer((_) async => WalletTestModel.uniqueId);
          when(_mockWalletApi.runSearchQuery(any))
              .thenAnswer((_) async => WalletTestModel.tWalletListEmpty);
          when(_mockWalletApi.createWallet(any, any))
              .thenAnswer((_) async => _tM4eToMTNTransaction.to.id);
          when(_mockTransactionApi.initiateTransaction(any))
              .thenAnswer((_) async => _m4eToMTNTransactionResult);

          // act
          final _result = await _m4eTransactionService.initiateTransaction(
              _userId, _tM4eToMTNTransaction);

          // assert
          verify(_mockWalletApi
              .queryUserWallets(_composedMobileMoneyWalletSearchForm));
          verify(_mockWalletApi.runSearchQuery(WalletTestModel.uniqueId));
          verify(_mockWalletApi.createWallet(
              _userId, _tM4eToMTNTransaction.to.toWalletFormType()));
          verify(
              _mockTransactionApi.initiateTransaction(_tM4eToMTNTransaction));

          expect(_result, equals(_m4eToMTNTransactionResult));
        });
      });
    });

    group('isDisconnected |', () {
      setUp(() {
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => false);
      });

      test(
          'should throw [ServerException] with [NO_INTERNET_CONNECTION] error code',
          () async {
        // arrange
        when(_mockTransactionApi.initiateTransaction(any))
            .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

        // act
        final _call = _m4eTransactionService.initiateTransaction;

        // assert
        expect(() => _call(_fromUuId, _tM4eToM4eTransaction),
            throwsA(TypeMatcher<M4eServerException>()));
      });
    });
  });

  group('getUserTransactions | ', () {
    test('should verify device connectivity status', () async {
      // arrange
      when(_mockDataConnectionChecker.hasConnection)
          .thenAnswer((_) async => true);
      when(_mockTransactionApi.getUserTransactions(any))
          .thenAnswer((_) async => _transactionList);

      // act
      await _m4eTransactionService.getUserTransactions(_userId);

      // assert
      verify(_mockDataConnectionChecker.hasConnection);
    });

    group('isConnected', () {
      setUp(() {
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
      });

      test(
          'should call TransactionApi to get user-transaction-logs with userId',
          () async {
        // arrange
        when(_mockTransactionApi.getUserTransactions(any)).thenAnswer(
            (_) async => M4eTransactionList(transactions: _transactionLogs));

        // act
        await _m4eTransactionService.getUserTransactions(_userId);

        // assert
        verify(_mockTransactionApi.getUserTransactions(_userId));
      });

      test('should return TransactionLogList on success', () async {
        // arrange
        when(_mockTransactionApi.getUserTransactions(any)).thenAnswer(
            (_) async => M4eTransactionList(transactions: _transactionLogs));

        // act
        final _result =
            await _m4eTransactionService.getUserTransactions(_userId);

        // assert
        verify(_mockTransactionApi.getUserTransactions(_userId));
        expect(_result,
            equals(M4eTransactionList(transactions: _transactionLogs)));
      });
    });

    group('isDisconnected', () {
      setUp(() {
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => false);

        test(
            'should throw a ServerException when no internet connection is detected',
            () async {
          // arrange
          when(_mockTransactionApi.getUserTransactions(any))
              .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

          // act
          final _call = _m4eTransactionService.getUserTransactions;

          // assert
          expect(
              () => _call(_userId), throwsA(TypeMatcher<M4eServerException>()));
        });
      });
    });
  });
}
