import 'package:data_connection_checker/data_connection_checker.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:m4e_sdk/src/core/core.dart';
import 'package:m4e_sdk/src/data/api/apis.dart';
import 'package:m4e_sdk/src/data/services/services.dart';
import 'package:m4e_sdk/src/m4e_forms/m4e_forms.dart';
import 'package:m4e_sdk/src/models/models.dart';
import 'package:matcher/matcher.dart';
import 'package:mockito/mockito.dart';

class MockUserApi extends Mock implements UserApi {}

class MockAuthApi extends Mock implements AuthApi {}

class MockDataConnectionChecker extends Mock implements DataConnectionChecker {}

void main() {
  M4eUserService _m4eUserService;

  MockUserApi _mockUserApi;
  MockAuthApi _mockAuthApi;
  MockDataConnectionChecker _mockDataConnectionChecker;

  setUp(() {
    _mockUserApi = MockUserApi();
    _mockAuthApi = MockAuthApi();
    _mockDataConnectionChecker = MockDataConnectionChecker();

    _m4eUserService = M4eUserService(
      userApi: _mockUserApi,
      authApi: _mockAuthApi,
      connectionChecker: _mockDataConnectionChecker,
    );
  });

  final _uniqueId =
      M4eUniqueId.fromUniqueString('46e87882-4ff8-4770-9846-870fbbcc35e0');

  final _authToken = M4eAuthToken(id: _uniqueId, token: 'test token');

  final _authCredential = M4eAuthPhoneNumberForm(
      id: _uniqueId,
      phoneNumber: M4ePhoneNumber(country: 'GHANA', number: 'test'));

  final _phoneNumber =
      M4ePhoneNumber(country: 'GHANA', number: '+233708788110');

  final _m4eUser = M4eUser(
    id: _uniqueId,
    shortName: 'test',
    firstName: 'test',
    lastName: 'test',
    phoneNumber: _phoneNumber,
    joined: DateTime.now(),
  );

  final _m4eUserForm = M4eUserForm(
    firstName: 'test',
    lastName: 'test',
    phoneNumber: _phoneNumber,
  );

  group('m4e-user-test getM4eUser', () {
    test('should verify device connectivity status', () async {
      // arrange
      when(_mockAuthApi.getUserCredentialsFromCache())
          .thenAnswer((_) async => _authCredential);
      when(_mockDataConnectionChecker.hasConnection)
          .thenAnswer((_) async => true);

      // act
      await _m4eUserService.getM4eUser();

      // assert
      verify(_mockDataConnectionChecker.hasConnection);
    });

    group('isConnected', () {
      setUp(() {
        when(_mockAuthApi.getUserCredentialsFromCache())
            .thenAnswer((_) async => _authCredential);
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
      });

      test('should get [AuthCredential] from cache', () async {
        // act
        await _m4eUserService.getM4eUser();

        // assert
        verify(_mockAuthApi.getUserCredentialsFromCache());
      });

      test('should return [M4eUser] if server call is successful', () async {
        // arrange
        when(_mockUserApi.getM4eUser(any)).thenAnswer((_) async => _m4eUser);

        // act
        final _result = await _m4eUserService.getM4eUser();

        // assert
        expect(_result, equals(_m4eUser));
      });

      test('should throw a [ServerException] if server call is unsuccessful',
          () async {
        // arrange
        when(_mockUserApi.getM4eUser(any))
            .thenThrow(M4eExceptionMessages.kUndefinedException);

        // act
        final _call = _m4eUserService.getM4eUser;

        // assert
        expect(() => _call(), throwsA(TypeMatcher<M4eServerException>()));
      });

      test('should throw an [AuthException] if no credential is returned',
          () async {
        // arrange
        when(_mockAuthApi.getUserCredentialsFromCache())
            .thenThrow(M4eExceptionMessages.kNoAuthCredential);

        // act
        final _call = _m4eUserService.getM4eUser;

        // assert
        expect(() => _call(), throwsA(TypeMatcher<M4eAuthException>()));
      });
    });

    group('isDisconnected', () {
      setUp(() {
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => false);
      });

      test(
          'should throw a ServerException when no internet connection is detected',
          () async {
        // arrange
        when(_mockAuthApi.getAuthTokenFromCache())
            .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

        // act
        final _call = _m4eUserService.getM4eUser;

        // assert
        expect(() => _call(), throwsA(TypeMatcher<M4eServerException>()));
      });
    });
  });

  group('m4e-user-test registerUser', () {
    test('should verify device connectivity status', () async {
      // arrange
      when(_mockAuthApi.getAuthTokenFromCache())
          .thenAnswer((_) async => _authToken);
      when(_mockDataConnectionChecker.hasConnection)
          .thenAnswer((_) async => true);

      // act
      await _m4eUserService.createM4eUser(_m4eUserForm);

      // assert
      verify(_mockDataConnectionChecker.hasConnection);
    });

    group('isConnected', () {
      setUp(() {
        when(_mockAuthApi.getAuthTokenFromCache())
            .thenAnswer((_) async => _authToken);
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => true);
      });

      test('should get [AuthToken] from cache', () async {
        // act
        await _m4eUserService.createM4eUser(_m4eUserForm);

        // assert
        verify(_mockAuthApi.getAuthTokenFromCache());
      });

      test('should return [UniqueId] if server call is successful', () async {
        // arrange
        when(_mockUserApi.createM4eUser(any, any))
            .thenAnswer((_) async => _uniqueId);

        // act
        final _result = await _m4eUserService.createM4eUser(_m4eUserForm);

        // assert
        expect(_result, equals(_uniqueId));
      });

      test('should throw a [ServerException] if server call is unsuccessful',
          () async {
        // arrange
        when(_mockUserApi.createM4eUser(any, any))
            .thenThrow(M4eExceptionMessages.kUndefinedException);

        // act
        final _call = _m4eUserService.createM4eUser;

        // assert
        expect(() => _call(_m4eUserForm),
            throwsA(TypeMatcher<M4eServerException>()));
      });

      test('should throw an [AuthException] if no user is authenticated',
          () async {
        // arrange
        when(_mockAuthApi.getAuthTokenFromCache())
            .thenThrow(M4eExceptionMessages.kUnauthorized);

        // act
        final _call = _m4eUserService.createM4eUser;

        // assert
        expect(() => _call(_m4eUserForm),
            throwsA(TypeMatcher<M4eAuthException>()));
      });
    });

    group('isDisconnected', () {
      setUp(() {
        when(_mockDataConnectionChecker.hasConnection)
            .thenAnswer((_) async => false);
      });

      test(
          'should throw a ServerException when no internet connection is detected',
          () async {
        // arrange
        when(_mockAuthApi.getAuthTokenFromCache())
            .thenThrow(M4eExceptionMessages.kNoInternetConnectionException);

        // act
        final _call = _m4eUserService.createM4eUser;

        // assert
        expect(() => _call(_m4eUserForm),
            throwsA(TypeMatcher<M4eServerException>()));
      });
    });
  });
}
